@@ -1,23 +1,17 @@
-from __future__ import unicode_literals
-
 import base64
 import logging
 from decimal import Decimal
 
-from six import string_types
-
 from .ewsdatetime import EWSDateTime
 from .properties import EWSElement
 from .util import create_element, add_xml_child, get_xml_attrs, get_xml_attr, set_xml_value, value_to_xml_text, \
-    xml_text_to_value, is_iterable, TNS
+    xml_text_to_value, is_iterable, safe_b64decode, TNS
 
 log = logging.getLogger(__name__)
 
 
 class ExtendedProperty(EWSElement):
-    """
-    MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/extendedproperty
-    """
+    """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/extendedproperty"""
     ELEMENT_NAME = 'ExtendedProperty'
 
     # Enum values: https://docs.microsoft.com/en-us/dotnet/api/exchangewebservices.distinguishedpropertysettype
@@ -101,7 +95,7 @@ class ExtendedProperty(EWSElement):
             # Allow to set attributes without keyword
             kwargs = dict(zip(self._slots_keys(), args))
         self.value = kwargs.pop('value')
-        super(ExtendedProperty, self).__init__(**kwargs)
+        super().__init__(**kwargs)
 
     @classmethod
     def validate_cls(cls):
@@ -201,6 +195,7 @@ class ExtendedProperty(EWSElement):
         # Returns whether an 'ExtendedProperty' element matches the definition for this class. Extended property fields
         # do not have a name, so we must match on the cls.property_* attributes to match a field in the request with a
         # field in the response.
+        # TODO: Rewrite to take advantage of exchangelib.properties.ExtendedFieldURI
         extended_field_uri = elem.find('{%s}ExtendedFieldURI' % TNS)
         cls_props = cls.properties_map()
         elem_props = {k: extended_field_uri.get(k) for k in cls_props.keys()}
@@ -229,15 +224,15 @@ class ExtendedProperty(EWSElement):
         if cls.is_array_type():
             values = elem.find('{%s}Values' % TNS)
             if cls.is_binary_type():
-                return [base64.b64decode(val) for val in get_xml_attrs(values, '{%s}Value' % TNS)]
+                return [safe_b64decode(val) for val in get_xml_attrs(values, '{%s}Value' % TNS)]
             return [
                 xml_text_to_value(value=val, value_type=python_type)
                 for val in get_xml_attrs(values, '{%s}Value' % TNS)
             ]
         if cls.is_binary_type():
-            return base64.b64decode(get_xml_attr(elem, '{%s}Value' % TNS))
+            return safe_b64decode(get_xml_attr(elem, '{%s}Value' % TNS))
         extended_field_value = xml_text_to_value(value=get_xml_attr(elem, '{%s}Value' % TNS), value_type=python_type)
-        if python_type == string_types[0] and not extended_field_value:
+        if python_type == str and not extended_field_value:
             # For string types, we want to return the empty string instead of None if the element was
             # actually found, but there was no XML value. For other types, it would be more problematic
             # to make that distinction, e.g. return False for bool, 0 for int, etc.
@@ -267,7 +262,7 @@ class ExtendedProperty(EWSElement):
 
     @classmethod
     def property_tag_as_int(cls):
-        if isinstance(cls.property_tag, string_types):
+        if isinstance(cls.property_tag, str):
             return int(cls.property_tag, base=16)
         return cls.property_tag
 
@@ -283,7 +278,7 @@ class ExtendedProperty(EWSElement):
             'ApplicationTime': Decimal,
             'Binary': bytes,
             'Boolean': bool,
-            'CLSID': string_types[0],
+            'CLSID': str,
             'Currency': int,
             'Double': Decimal,
             'Float': Decimal,
@@ -291,7 +286,7 @@ class ExtendedProperty(EWSElement):
             'Long': int,
             'Short': int,
             'SystemTime': EWSDateTime,
-            'String': string_types[0],
+            'String': str,
         }[base_type]
 
     @classmethod
@@ -310,9 +305,21 @@ class ExtendedProperty(EWSElement):
 class ExternId(ExtendedProperty):
     """This is a custom extended property defined by us. It's useful for synchronization purposes, to attach a unique ID
     from an external system.
+
     """
     property_set_id = 'c11ff724-aa03-4555-9952-8fa248a11c3e'  # This is arbitrary. We just want a unique UUID.
     property_name = 'External ID'
     property_type = 'String'
 
     __slots__ = tuple()
+
+
+class Flag(ExtendedProperty):
+    """This property returns None for Not Flagged messages, 1 for Completed messages and 2 for Flagged messages.
+
+    For a description of each status, see:
+    https://docs.microsoft.com/en-us/openspecs/exchange_server_protocols/ms-oxoflag/eda9fd25-6407-4cec-9e62-26e4f9d6a098
+
+    """
+    property_tag = 0x1090
+    property_type = 'Integer'
