@@ -1,60 +1,66 @@
-from __future__ import unicode_literals
-
 from base64 import b64decode
 from codecs import BOM_UTF8
 from collections import OrderedDict
 import datetime
 from decimal import Decimal
+from functools import wraps
 import io
 import itertools
 import logging
 import re
 import socket
+from threading import get_ident
 import time
-import xml.sax.handler
+from urllib.parse import urlparse
+import xml.sax.handler  # nosec
 
-# Import _etree via defusedxml instead of directly from lxml.etree, to silence overly strict linters
-from defusedxml.lxml import parse, tostring, GlobalParserTLS, RestrictedElement, _etree
+import lxml.etree  # nosec
 from defusedxml.expatreader import DefusedExpatParser
 from defusedxml.sax import _InputSource
-from future.backports.misc import get_ident
-from future.moves.urllib.parse import urlparse
-from future.utils import PY2
 import isodate
+from oauthlib.oauth2 import TokenExpiredError
 from pygments import highlight
 from pygments.lexers.html import XmlLexer
 from pygments.formatters.terminal import TerminalFormatter
 import requests.exceptions
-from six import text_type, string_types
 
 from .errors import TransportError, RateLimitError, RedirectError, RelativeRedirect, CASError, UnauthorizedError, \
     ErrorInvalidSchemaVersionForMailboxVersion
 
-time_func = time.time if PY2 else time.monotonic
 log = logging.getLogger(__name__)
 
 
-class ParseError(_etree.ParseError):
+def require_account(f):
+    @wraps(f)
+    def wrapper(self, *args, **kwargs):
+        if not self.account:
+            raise ValueError('%s must have an account' % self.__class__.__name__)
+        return f(self, *args, **kwargs)
+    return wrapper
+
+
+def require_id(f):
+    @wraps(f)
+    def wrapper(self, *args, **kwargs):
+        if not self.account:
+            raise ValueError('%s must have an account' % self.__class__.__name__)
+        if not self.id:
+            raise ValueError('%s must have an ID' % self.__class__.__name__)
+        return f(self, *args, **kwargs)
+    return wrapper
+
+
+class ParseError(lxml.etree.ParseError):
     """Used to wrap lxml ParseError in our own class"""
     pass
 
 
 class ElementNotFound(Exception):
     def __init__(self, msg, data):
-        super(ElementNotFound, self).__init__(msg)
+        super().__init__(msg)
         self.data = data
 
 
-class PickleMixIn(object):
-    if PY2:
-        def __getstate__(self):
-            return {k: getattr(self, k) for k in self.__slots__}
-
-        def __setstate__(self, state):
-            for k in self.__slots__:
-                setattr(self, k, state.get(k))
-
-
 # Regex of UTF-8 control characters that are illegal in XML 1.0 (and XML 1.1)
 _ILLEGAL_XML_CHARS_RE = re.compile('[\x00-\x08\x0b\x0c\x0e-\x1F\uD800-\uDFFF\uFFFE\uFFFF]')
 
@@ -63,6 +69,9 @@ SOAPNS = 'http://schemas.xmlsoap.org/soap/envelope/'
 MNS = 'http://schemas.microsoft.com/exchange/services/2006/messages'
 TNS = 'http://schemas.microsoft.com/exchange/services/2006/types'
 ENS = 'http://schemas.microsoft.com/exchange/services/2006/errors'
+AUTODISCOVER_BASE_NS = 'http://schemas.microsoft.com/exchange/autodiscover/responseschema/2006'
+AUTODISCOVER_REQUEST_NS = 'http://schemas.microsoft.com/exchange/autodiscover/outlook/requestschema/2006'
+AUTODISCOVER_RESPONSE_NS = 'http://schemas.microsoft.com/exchange/autodiscover/outlook/responseschema/2006a'
 
 ns_translation = OrderedDict([
     ('s', SOAPNS),
@@ -70,7 +79,7 @@ ns_translation = OrderedDict([
     ('t', TNS),
 ])
 for item in ns_translation.items():
-    _etree.register_namespace(*item)
+    lxml.etree.register_namespace(*item)
 
 
 def is_iterable(value, generators_allowed=False):
@@ -78,12 +87,16 @@ def is_iterable(value, generators_allowed=False):
     callers don't necessarily guarantee that they only iterate the value once. Take care to not match string types and
     bytes.
 
-    :param value: any type of object
-    :param generators_allowed: if True, generators will be treated as iterable
-    :return: True or False
+    Args:
+      value: any type of object
+      generators_allowed: if True, generators will be treated as iterable (Default value = False)
+
+    Returns:
+      True or False
+
     """
     if generators_allowed:
-        if not isinstance(value, string_types + (bytes,)) and hasattr(value, '__iter__'):
+        if not isinstance(value, (bytes, str)) and hasattr(value, '__iter__'):
             return True
     else:
         if isinstance(value, (tuple, list, set)):
@@ -92,7 +105,13 @@ def is_iterable(value, generators_allowed=False):
 
 
 def chunkify(iterable, chunksize):
-    """Splits an iterable into chunks of size ``chunksize``. The last chunk may be smaller than ``chunksize``."""
+    """Splits an iterable into chunks of size ``chunksize``. The last chunk may be smaller than ``chunksize``.
+
+    Args:
+      iterable:
+      chunksize:
+
+    """
     from .queryset import QuerySet
     if hasattr(iterable, '__getitem__') and not isinstance(iterable, QuerySet):
         # tuple, list. QuerySet has __getitem__ but that evaluates the entire query greedily. We don't want that here.
@@ -111,7 +130,12 @@ def chunkify(iterable, chunksize):
 
 
 def peek(iterable):
-    """Checks if an iterable is empty and returns status and the rewinded iterable"""
+    """Checks if an iterable is empty and returns status and the rewinded iterable
+
+    Args:
+      iterable:
+
+    """
     from .queryset import QuerySet
     if isinstance(iterable, QuerySet):
         # QuerySet has __len__ but that evaluates the entire query greedily. We don't want that here. Instead, peek()
@@ -130,12 +154,19 @@ def peek(iterable):
 
 
 def xml_to_str(tree, encoding=None, xml_declaration=False):
-    """Serialize an XML tree. Returns unicode if 'encoding' is None. Otherwise, we return encoded 'bytes'."""
+    """Serialize an XML tree. Returns unicode if 'encoding' is None. Otherwise, we return encoded 'bytes'.
+
+    Args:
+      tree:
+      encoding:  (Default value = None)
+      xml_declaration:  (Default value = False)
+
+    """
     if xml_declaration and not encoding:
         raise ValueError("'xml_declaration' is not supported when 'encoding' is None")
     if encoding:
-        return tostring(tree, encoding=encoding, xml_declaration=True)
-    return tostring(tree, encoding=text_type, xml_declaration=False)
+        return lxml.etree.tostring(tree, encoding=encoding, xml_declaration=True)
+    return lxml.etree.tostring(tree, encoding=str, xml_declaration=False)
 
 
 def get_xml_attr(tree, name):
@@ -154,12 +185,12 @@ def value_to_xml_text(value):
     from .ewsdatetime import EWSTimeZone, EWSDateTime, EWSDate
     from .indexed_properties import PhoneNumber, EmailAddress
     from .properties import Mailbox, Attendee, ConversationId
-    if isinstance(value, string_types):
+    if isinstance(value, str):
         return safe_xml_value(value)
     if isinstance(value, bool):
         return '1' if value else '0'
     if isinstance(value, (int, Decimal)):
-        return text_type(value)
+        return str(value)
     if isinstance(value, datetime.time):
         return value.isoformat()
     if isinstance(value, EWSTimeZone):
@@ -190,7 +221,7 @@ def xml_text_to_value(value, value_type):
         Decimal: Decimal,
         datetime.timedelta: isodate.parse_duration,
         EWSDateTime: EWSDateTime.from_string,
-        string_types[0]: lambda v: v
+        str: lambda v: v
     }[value_type](value)
 
 
@@ -199,9 +230,9 @@ def set_xml_value(elem, value, version):
     from .fields import FieldPath, FieldOrder
     from .properties import EWSElement
     from .version import Version
-    if isinstance(value, string_types + (bool, bytes, int, Decimal, datetime.time, EWSDate, EWSDateTime)):
+    if isinstance(value, (str, bool, bytes, int, Decimal, datetime.time, EWSDate, EWSDateTime)):
         elem.text = value_to_xml_text(value)
-    elif isinstance(value, RestrictedElement):
+    elif isinstance(value, _element_class):
         elem.append(value)
     elif is_iterable(value, generators_allowed=True):
         for v in value:
@@ -211,9 +242,9 @@ def set_xml_value(elem, value, version):
                 if not isinstance(version, Version):
                     raise ValueError("'version' %r must be a Version instance" % version)
                 elem.append(v.to_xml(version=version))
-            elif isinstance(v, RestrictedElement):
+            elif isinstance(v, _element_class):
                 elem.append(v)
-            elif isinstance(v, string_types):
+            elif isinstance(v, str):
                 add_xml_child(elem, 't:String', v)
             else:
                 raise ValueError('Unsupported type %s for list element %s on elem %s' % (type(v), v, elem))
@@ -221,7 +252,7 @@ def set_xml_value(elem, value, version):
         elem.append(value.to_xml())
     elif isinstance(value, EWSElement):
         if not isinstance(version, Version):
-            raise ValueError("'version' %s must be a Version instance" % version)
+            raise ValueError("'version' %r must be a Version instance" % version)
         elem.append(value.to_xml(version=version))
     else:
         raise ValueError('Unsupported type %s for value %s on elem %s' % (type(value), value, elem))
@@ -229,7 +260,7 @@ def set_xml_value(elem, value, version):
 
 
 def safe_xml_value(value, replacement='?'):
-    return text_type(_ILLEGAL_XML_CHARS_RE.sub(replacement, value))
+    return _ILLEGAL_XML_CHARS_RE.sub(replacement, value)
 
 
 def create_element(name, attrs=None, nsmap=None):
@@ -238,12 +269,11 @@ def create_element(name, attrs=None, nsmap=None):
     if ':' in name:
         ns, name = name.split(':')
         name = '{%s}%s' % (ns_translation[ns], name)
-    elem = RestrictedElement(nsmap=nsmap)
+    elem = _forgiving_parser.makeelement(name, nsmap=nsmap)
     if attrs:
         # Try hard to keep attribute order, to ensure deterministic output. This simplifies testing.
         for k, v in attrs.items():
             elem.set(k, v)
-    elem.tag = name
     return elem
 
 
@@ -256,7 +286,9 @@ def add_xml_child(tree, name, value):
 class StreamingContentHandler(xml.sax.handler.ContentHandler):
     """A SAX content handler that returns a character data for a single element back to the parser. The parser must have
     a 'buffer' attribute we can append data to.
+
     """
+
     def __init__(self, parser, ns, element_name):
         xml.sax.handler.ContentHandler.__init__(self)
         self._parser = parser
@@ -289,16 +321,30 @@ def prepare_input_source(source):
     return source
 
 
+def safe_b64decode(data):
+    # Incoming base64-encoded data is not always padded to a multiple of 4. Python's parser is more strict and requires
+    # padding. Add padding if it's needed.
+    overflow = len(data) % 4
+    if overflow:
+        if isinstance(data, str):
+            padding = '=' * (4 - overflow)
+        else:
+            padding = b'=' * (4 - overflow)
+        data += padding
+    return b64decode(data)
+
+
 class StreamingBase64Parser(DefusedExpatParser):
     """A SAX parser that returns a generator of base64-decoded character content"""
+
     def __init__(self, *args, **kwargs):
         DefusedExpatParser.__init__(self, *args, **kwargs)
         self._namespaces = True
         self.buffer = None
         self.element_found = None
 
-    def parse(self, source):
-        raw_source = source.raw
+    def parse(self, r):
+        raw_source = r.raw
         # Like upstream but yields the return value of self.feed()
         raw_source = prepare_input_source(raw_source)
         self.prepareParser(raw_source)
@@ -313,15 +359,13 @@ class StreamingBase64Parser(DefusedExpatParser):
             for data in self.feed(buffer):
                 yield data
             buffer = file.read(self._bufsize)
+        # Any remaining data in self.buffer should be padding chars now
         self.buffer = None
-        source.close()
+        r.close()  # Release memory
         self.close()
         if not self.element_found:
-            if PY2:
-                data = b''.join(collected_data)
-            else:
-                data = bytes(collected_data)
-            raise ElementNotFound('The element to be streamed from was not found', data=data)
+            data = bytes(collected_data)
+            raise ElementNotFound('The element to be streamed from was not found', data=bytes(data))
 
     def feed(self, data, isFinal=0):
         # Like upstream, but yields the current content of the character buffer
@@ -332,7 +376,7 @@ class StreamingBase64Parser(DefusedExpatParser):
         remainder = ''
         for data in self.buffer:
             available = len(remainder) + len(data)
-            overflow = available % 4
+            overflow = available % 4  # Make sure we always decode a multiple of 4
             if remainder:
                 data = (remainder + data)
                 remainder = ''
@@ -343,26 +387,26 @@ class StreamingBase64Parser(DefusedExpatParser):
         self.buffer = [remainder] if remainder else []
 
 
-class ForgivingParser(GlobalParserTLS):
-    parser_config = {
-        'resolve_entities': False,
-        'recover': True,  # This setting is non-default
-    }
-
-
-_forgiving_parser = ForgivingParser()
+_forgiving_parser = lxml.etree.XMLParser(
+    resolve_entities=False,  # This setting is recommended by lxml for safety
+    recover=True,  # This setting is non-default
+    huge_tree=True,  # This setting enables parsing huge attachments, mime_content and other large data
+)
+_element_class = _forgiving_parser.makeelement('x').__class__
 
 
 class BytesGeneratorIO(io.RawIOBase):
     """A BytesIO that can produce bytes from a streaming HTTP request. Expects r.iter_content() as input
     lxml tries to be smart by calling `getvalue` when present, assuming that the entire string is in memory.
     Omitting `getvalue` forces lxml to stream the request through `read` avoiding the memory duplication.
+
     """
+
     def __init__(self, bytes_generator):
         self._bytes_generator = bytes_generator
         self._next = bytearray()
         self._tell = 0
-        super(BytesGeneratorIO, self).__init__()
+        super().__init__()
 
     def readable(self):
         return not self.closed
@@ -398,7 +442,7 @@ class BytesGeneratorIO(io.RawIOBase):
     def close(self):
         if not self.closed:
             self._bytes_generator.close()
-        super(BytesGeneratorIO, self).close()
+        super().close()
 
 
 def to_xml(bytes_content):
@@ -408,25 +452,24 @@ def to_xml(bytes_content):
         stream = io.BytesIO(bytes_content)
     else:
         stream = BytesGeneratorIO(bytes_content)
-    forgiving_parser = _forgiving_parser.getDefaultParser()
     try:
-        return parse(stream, parser=forgiving_parser)
+        res = lxml.etree.parse(stream, parser=_forgiving_parser)  # nosec
     except AssertionError as e:
         raise ParseError(e.args[0], '<not from file>', -1, 0)
-    except _etree.ParseError as e:
+    except lxml.etree.ParseError as e:
         if hasattr(e, 'position'):
             e.lineno, e.offset = e.position
         if not e.lineno:
-            raise ParseError(text_type(e), '<not from file>', e.lineno, e.offset)
+            raise ParseError(str(e), '<not from file>', e.lineno, e.offset)
         try:
             stream.seek(0)
             offending_line = stream.read().splitlines()[e.lineno - 1]
         except (IndexError, io.UnsupportedOperation):
-            raise ParseError(text_type(e), '<not from file>', e.lineno, e.offset)
+            raise ParseError(str(e), '<not from file>', e.lineno, e.offset)
         else:
             offending_excerpt = offending_line[max(0, e.offset - 20):e.offset + 20]
-            msg = '%s\nOffending text: [...]%s[...]' % (text_type(e), offending_excerpt)
-            raise ParseError(msg, e.lineno, e.offset)
+            msg = '%s\nOffending text: [...]%s[...]' % (str(e), offending_excerpt)
+            raise ParseError(msg, '<not from file>', e.lineno, e.offset)
     except TypeError:
         try:
             stream.seek(0)
@@ -434,28 +477,44 @@ def to_xml(bytes_content):
             pass
         raise ParseError('This is not XML: %r' % stream.read(), '<not from file>', -1, 0)
 
+    if res.getroot() is None:
+        try:
+            stream.seek(0)
+            msg = 'No root element found: %r' % stream.read()
+        except (IndexError, io.UnsupportedOperation):
+            msg = 'No root element found'
+        raise ParseError(msg, '<not from file>', -1, 0)
+    return res
+
+
+def is_xml(text, expected_prefix=b'<?xml'):
+    """Lightweight test if response is an XML doc. It's better to be fast than correct here.
+
+    Args:
+      text: The string to check
+      expected_prefix: What to search for in the start if the string
 
-def is_xml(text):
-    """
-    Helper function. Lightweight test if response is an XML doc
     """
     # BOM_UTF8 is an UTF-8 byte order mark which may precede the XML from an Exchange server
     bom_len = len(BOM_UTF8)
+    prefix_len = len(expected_prefix)
     if text[:bom_len] == BOM_UTF8:
-        return text[bom_len:bom_len + 5] == b'<?xml'
-    return text[:5] == b'<?xml'
+        prefix = text[bom_len:bom_len + prefix_len]
+    else:
+        prefix = text[:prefix_len]
+    return prefix == expected_prefix
 
 
 class PrettyXmlHandler(logging.StreamHandler):
     """A steaming log handler that prettifies log statements containing XML when output is a terminal"""
     @staticmethod
     def parse_bytes(xml_bytes):
-        return parse(io.BytesIO(xml_bytes))
+        return lxml.etree.parse(io.BytesIO(xml_bytes), parser=_forgiving_parser)  # nosec
 
     @classmethod
     def prettify_xml(cls, xml_bytes):
         # Re-formats an XML document to a consistent style
-        return tostring(
+        return lxml.etree.tostring(
             cls.parse_bytes(xml_bytes),
             xml_declaration=True,
             encoding='utf-8',
@@ -472,6 +531,10 @@ class PrettyXmlHandler(logging.StreamHandler):
            * This is a DEBUG message
            * We're outputting to a terminal
            * The log message args is a dict containing keys starting with 'xml_' and values as bytes
+
+        Args:
+          record:
+
         """
         if record.levelno == logging.DEBUG and self.is_tty() and isinstance(record.args, dict):
             for key, value in record.args.items():
@@ -482,14 +545,11 @@ class PrettyXmlHandler(logging.StreamHandler):
                 if not is_xml(value):
                     continue
                 try:
-                    if PY2:
-                        record.args[key] = self.highlight_xml(self.prettify_xml(value)).encode('utf-8')
-                    else:
-                        record.args[key] = self.highlight_xml(self.prettify_xml(value))
+                    record.args[key] = self.highlight_xml(self.prettify_xml(value))
                 except Exception as e:
                     # Something bad happened, but we don't want to crash the program just because logging failed
                     print('XML highlighting failed: %s' % e)
-        return super(PrettyXmlHandler, self).emit(record)
+        return super().emit(record)
 
     def is_tty(self):
         # Check if we're outputting to a terminal
@@ -501,12 +561,13 @@ class PrettyXmlHandler(logging.StreamHandler):
 
 class AnonymizingXmlHandler(PrettyXmlHandler):
     """A steaming log handler that prettifies and anonymizes log statements containing XML when output is a terminal"""
+
     def __init__(self, forbidden_strings, *args, **kwargs):
         self.forbidden_strings = forbidden_strings
-        super(AnonymizingXmlHandler, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
 
     def parse_bytes(self, xml_bytes):
-        root = parse(io.BytesIO(xml_bytes))
+        root = lxml.etree.parse(io.BytesIO(xml_bytes), parser=_forgiving_parser)  # nosec
         for elem in root.iter():
             for attr in set(elem.keys()) & {'RootItemId', 'ItemId', 'Id', 'RootItemChangeKey', 'ChangeKey'}:
                 elem.set(attr, 'DEADBEEF=')
@@ -515,23 +576,27 @@ class AnonymizingXmlHandler(PrettyXmlHandler):
         return root
 
 
-class DummyRequest(object):
+class DummyRequest:
     def __init__(self, headers):
         self.headers = headers
 
 
-class DummyResponse(object):
-    def __init__(self, url, headers, request_headers, content=b''):
-        self.status_code = 503
+class DummyResponse:
+    def __init__(self, url, headers, request_headers, content=b'', status_code=503, history=None):
+        self.status_code = status_code
         self.url = url
         self.headers = headers
         self.content = content
         self.text = content.decode('utf-8', errors='ignore')
         self.request = DummyRequest(headers=request_headers)
+        self.history = history
 
     def iter_content(self):
         return self.content
 
+    def close(self):
+        pass
+
 
 def get_domain(email):
     try:
@@ -542,7 +607,7 @@ def get_domain(email):
 
 def split_url(url):
     parsed_url = urlparse(url)
-    # Use netloc instead og hostname since hostname is None if URL is relative
+    # Use netloc instead of hostname since hostname is None if URL is relative
     return parsed_url.scheme == 'https', parsed_url.netloc.lower(), parsed_url.path
 
 
@@ -577,14 +642,12 @@ def get_redirect_url(response, allow_relative=True, require_relative=False):
     return redirect_url
 
 
-MAX_REDIRECTS = 5  # Define a max redirection count. We don't want to be sent into an endless redirect loop
+RETRY_WAIT = 10  # Seconds to wait before retry on connection errors
+MAX_REDIRECTS = 10  # Maximum number of URL redirects before we give up
 
 # A collection of error classes we want to handle as general connection errors
 CONNECTION_ERRORS = (requests.exceptions.ChunkedEncodingError, requests.exceptions.ConnectionError,
-                     requests.exceptions.Timeout, socket.timeout)
-if not PY2:
-    # Python2 does not have ConnectionResetError
-    CONNECTION_ERRORS += (ConnectionResetError,)  # noqa: F821
+                     requests.exceptions.Timeout, socket.timeout, ConnectionResetError)
 
 # A collection of error classes we want to handle as TLS verification errors
 TLS_ERRORS = (requests.exceptions.SSLError,)
@@ -597,8 +660,7 @@ except ImportError:
 
 
 def post_ratelimited(protocol, session, url, headers, data, allow_redirects=False, stream=False):
-    """
-    There are two error-handling policies implemented here: a fail-fast policy intended for stand-alone scripts which
+    """There are two error-handling policies implemented here: a fail-fast policy intended for stand-alone scripts which
     fails on all responses except HTTP 200. The other policy is intended for long-running tasks that need to respect
     rate-limiting errors from the server and paper over outages of up to 1 hour.
 
@@ -621,15 +683,22 @@ def post_ratelimited(protocol, session, url, headers, data, allow_redirects=Fals
 
     An additional resource on handling throttling policies and client back off strategies:
         https://docs.microsoft.com/en-us/exchange/client-developer/exchange-web-services/ews-throttling-in-exchange
+
+    Args:
+      protocol:
+      session:
+      url:
+      headers:
+      data:
+      allow_redirects:  (Default value = False)
+      stream:  (Default value = False)
+
     """
     thread_id = get_ident()
-    wait = 10  # seconds
+    wait = RETRY_WAIT  # Initial retry wait. We double the value on each retry
     retry = 0
     redirects = 0
-    # In Python 2, we want this to be a 'str' object so logging doesn't break (all formatting arguments are 'str').
-    # We activated 'unicode_literals' at the top of this file, so it would be a 'unicode' object unless we convert
-    # to 'str' explicitly. This is a no-op for Python 3.
-    log_msg = str('''\
+    log_msg = '''\
 Retry: %(retry)s
 Waited: %(wait)s
 Timeout: %(timeout)s
@@ -646,7 +715,7 @@ Request headers: %(request_headers)s
 Response headers: %(response_headers)s
 Request data: %(xml_request)s
 Response data: %(xml_response)s
-''')
+'''
     log_vals = dict(
         retry=retry,
         wait=wait,
@@ -665,44 +734,63 @@ Response data: %(xml_response)s
         xml_request=data,
         xml_response=None,
     )
+    t_start = time.monotonic()
     try:
         while True:
-            _back_off_if_needed(protocol.retry_policy.back_off_until)
+            backed_off = _back_off_if_needed(protocol.retry_policy.back_off_until)
+            if backed_off:
+                # We may have slept for a long time. Renew the session.
+                session = protocol.renew_session(session)
             log.debug('Session %s thread %s: retry %s timeout %s POST\'ing to %s after %ss wait', session.session_id,
                       thread_id, retry, protocol.TIMEOUT, url, wait)
-            d_start = time_func()
+            d_start = time.monotonic()
             # Always create a dummy response for logging purposes, in case we fail in the following
             r = DummyResponse(url=url, headers={}, request_headers=headers)
             try:
                 r = session.post(url=url, headers=headers, data=data, allow_redirects=False, timeout=protocol.TIMEOUT,
                                  stream=stream)
+            except TLS_ERRORS as e:
+                # Don't retry on TLS errors. They will most likely be persistent.
+                raise TransportError(str(e))
             except CONNECTION_ERRORS as e:
                 log.debug('Session %s thread %s: connection error POST\'ing to %s', session.session_id, thread_id, url)
                 r = DummyResponse(url=url, headers={'TimeoutException': e}, request_headers=headers)
+            except TokenExpiredError as e:
+                log.debug('Session %s thread %s: OAuth token expired; refreshing', session.session_id, thread_id)
+                r = DummyResponse(url=url, headers={'TokenExpiredError': e}, request_headers=headers, status_code=401)
+            except KeyError as e:
+                if e.args[0] != 'www-authenticate':
+                    raise
+                log.debug('Session %s thread %s: auth headers missing from %s', session.session_id, thread_id, url)
+                r = DummyResponse(url=url, headers={'KeyError': e}, request_headers=headers)
             finally:
                 log_vals.update(
                     retry=retry,
                     wait=wait,
                     session_id=session.session_id,
                     url=str(r.url),
-                    response_time=time_func() - d_start,
+                    response_time=time.monotonic() - d_start,
                     status_code=r.status_code,
                     request_headers=r.request.headers,
                     response_headers=r.headers,
                     xml_response='[STREAMING]' if stream else r.content,
                 )
             log.debug(log_msg, log_vals)
-            if _may_retry_on_error(r, protocol, wait):
+            if _need_new_credentials(response=r):
+                r.close()  # Release memory
+                session = protocol.refresh_credentials(session)
+                continue
+            total_wait = time.monotonic() - t_start
+            if _may_retry_on_error(response=r, retry_policy=protocol.retry_policy, wait=total_wait):
+                r.close()  # Release memory
                 log.info("Session %s thread %s: Connection error on URL %s (code %s). Cool down %s secs",
                          session.session_id, thread_id, r.url, r.status_code, wait)
-                time.sleep(wait)  # Increase delay for every retry
+                protocol.retry_policy.back_off(wait)
                 retry += 1
-                wait *= 2
-                session = protocol.renew_session(session)
+                wait *= 2  # Increase delay for every retry
                 continue
             if r.status_code in (301, 302):
-                if stream:
-                    r.close()
+                r.close()  # Release memory
                 url, redirects = _redirect_or_fail(r, redirects, allow_redirects)
                 continue
             break
@@ -720,11 +808,7 @@ Response data: %(xml_response)s
         log.debug('Got status code %s but trying to parse content anyway', r.status_code)
     elif r.status_code != 200:
         protocol.retire_session(session)
-        try:
-            _raise_response_errors(r, protocol, log_msg, log_vals)  # Always raises an exception
-        finally:
-            if stream:
-                r.close()
+        _raise_response_errors(r, protocol, log_msg, log_vals)  # Always raises an exception
     log.debug('Session %s thread %s: Useful response from %s', session.session_id, thread_id, url)
     return r, session
 
@@ -736,9 +820,22 @@ def _back_off_if_needed(back_off_until):
         if sleep_secs > 0:
             log.warning('Server requested back off until %s. Sleeping %s seconds', back_off_until, sleep_secs)
             time.sleep(sleep_secs)
+            return True
+    return False
 
 
-def _may_retry_on_error(response, protocol, wait):
+def _may_retry_on_error(response, retry_policy, wait):
+    if response.status_code not in (301, 302, 401, 503):
+        # Don't retry if we didn't get a status code that we can hope to recover from
+        log.debug('No retry: wrong status code %s', response.status_code)
+        return False
+    if retry_policy.fail_fast:
+        log.debug('No retry: no fail-fast policy')
+        return False
+    if wait > retry_policy.max_wait:
+        # We lost patience. Session is cleaned up in outer loop
+        raise RateLimitError(
+            'Max timeout reached', url=response.url, status_code=response.status_code, total_wait=wait)
     # The genericerrorpage.htm/internalerror.asp is ridiculous behaviour for random outages. Redirect to
     # '/internalsite/internalerror.asp' or '/internalsite/initparams.aspx' is caused by e.g. TLS certificate
     # f*ckups on the Exchange server.
@@ -747,19 +844,16 @@ def _may_retry_on_error(response, protocol, wait):
             or (response.status_code == 302 and response.headers.get('location', '').lower() ==
                 '/ews/genericerrorpage.htm?aspxerrorpath=/ews/exchange.asmx') \
             or (response.status_code == 503):
-        if response.status_code not in (301, 302, 401, 503):
-            # Don't retry if we didn't get a status code that we can hope to recover from
-            return False
-        if protocol.retry_policy.fail_fast:
-            return False
-        if wait > protocol.retry_policy.max_wait:
-            # We lost patience. Session is cleaned up in outer loop
-            raise RateLimitError(
-                'Max timeout reached', url=response.url, status_code=response.status_code, total_wait=wait)
+        log.debug('Retry allowed: conditions met')
         return True
     return False
 
 
+def _need_new_credentials(response):
+    return response.status_code == 401 \
+        and response.headers.get('TokenExpiredError')
+
+
 def _redirect_or_fail(response, redirects, allow_redirects):
     # Retry with no delay. If we let requests handle redirects automatically, it would issue a GET to that
     # URL. We still want to POST.
@@ -791,7 +885,7 @@ def _raise_response_errors(response, protocol, log_msg, log_vals):
         raise TransportError('The service account is currently locked out')
     if response.status_code == 401 and protocol.retry_policy.fail_fast:
         # This is a login failure
-        raise UnauthorizedError('Wrong username or password for %s' % response.url)
+        raise UnauthorizedError('Invalid credentials for %s' % response.url)
     if 'TimeoutException' in response.headers:
         raise response.headers['TimeoutException']
     # This could be anything. Let higher layers handle this. Add full context for better debugging.
